// kleval.cc
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Compute the kernel matrix of a kernel.

#include <fst/fst.h>
#include "kernel/lib/kernel.h"
#include "kernel/lib/rational-kernel.h"

#include <tr1/unordered_map>
using std::tr1::unordered_map;
#include <fst/slist.h>
#include <string>
#include <vector>


DEFINE_int64(data_size, -1, "size of dataset to be evaluated [-1 means full dataset]");
DEFINE_bool(libsvm, false, "use libsvm precomputed kernel format");
DEFINE_int64(xmin,  0, "restrict to k(x,y) where x >= xmin");
DEFINE_int64(xmax, -1, "restrict to k(x,y) where x < xmax");
DEFINE_int64(ymin,  0, "restrict to k(x,y) where y >= ymin");
DEFINE_int64(ymax, -1, "restrict to k(x,y) where y < ymax");
DEFINE_bool(null, false, "null output");
DEFINE_bool(kar, false, "output kar file with cache instead of matrix in ascii");
DEFINE_bool(bin, false, "output kernel matrix in binary format (double)");

DEFINE_bool(normalize, false, "normalize features");
DEFINE_bool(log_domain, false, "compute features in log domain");
DEFINE_string(feature_fst, "", "dump fst encoding feature into specified file");


using namespace kernel;
using namespace fst;

namespace std {
namespace tr1 {

template<>
struct hash<basic_string<LogArc::Label> > {
  size_t operator()(const basic_string<LogArc::Label>& s) const {
    uint64 __h = 0;
    for (basic_string<LogArc::Label>::const_iterator __s = s.begin();
         __s != s.end();
         ++__s)
      __h = 5 * __h + *__s;
    return size_t(__h);
  }
};

}  // namespace tr1
}  // namespace std

template <class Arc>
void ListStrings(
    const Fst<Arc> &fst,
    vector<pair<basic_string<typename Arc::Label>, typename Arc::Weight> >
    *strings) {

  typedef typename Arc::Label Label;
  typedef basic_string<Label> String;
  typedef typename Arc::Weight Weight;
  typedef typename Arc::StateId StateId;
  typedef pair<StateId, pair<String, Weight> > Tuple;

  if (fst.Start() == kNoStateId)
    return;

  CHECK(fst.Properties(kODeterministic, true));


  vector<Tuple> queue;
  queue.push_back(make_pair(fst.Start(), make_pair(String(), Weight::One())));
  while (!queue.empty()) {
    Tuple tuple = queue.back();
    queue.pop_back();
    ArcIterator< Fst<Arc> > ait(fst, tuple.first);
    for (; !ait.Done(); ait.Next()) {
      Weight w = Times(tuple.second.second, ait.Value().weight);
      String s = tuple.second.first;
      Label label = ait.Value().olabel;
      if (label != 0)
        s.push_back(label);
      queue.push_back(make_pair(ait.Value().nextstate, make_pair(s, w)));
    }
    if (fst.Final(tuple.first) != Weight::Zero())
      strings->push_back(
          make_pair(tuple.second.first,
                    Times(tuple.second.second, fst.Final(tuple.first))));
  }
}

template <class Weight>
struct Compare {
  bool operator()(const pair<ssize_t, Weight> &x,
                const pair<ssize_t, Weight> &y) {
    return x.first < y.first;
  }
};


// TODO: need to fix constness of GetFunction
// TODO: enable saving the feature dictionary to a file
template <class Arc>
void ComputeFeatures(RationalKernel<Arc>* kernel,
                     ssize_t begin,
                     ssize_t end,
                     bool normalize,
                     bool log_domain) {
  typedef typename Arc::Label Label;
  typedef typename Arc::Weight Weight;
  typedef RationalFunction<Arc> KernelFunction;
  const KernelFunction *function = kernel->GetFunction();
  DataSet< Fst<Arc> > *dataset = kernel->Data();

  vector<basic_string<Label> > string_vector;
  unordered_map<basic_string<Label>, ssize_t> string_hash;


  for (int i = begin; i <= end; ++i) {
    // cerr << "#i\n";
    const VectorFst<Arc> *fst =
        function->GetFeature((*dataset)[i]);
    //     if (i == begin)
    //       fst->Write("foo.fst");
    vector<pair<basic_string<Label>, Weight> > strings;
    vector<pair<ssize_t, double> > features;
    ListStrings(*fst, &strings);
    double norm = 0.0F;
    //     if (normalize)
    //       norm = sqrt(function->Product(*fst, *fst));
    for (int j = 0; j < strings.size(); ++j) {
      ssize_t &index = string_hash[strings[j].first];
      if (index == 0) {
        string_vector.push_back(strings[j].first);
        index = string_vector.size();
      }
      //       {
      //         basic_string<Label> s = strings[j].first;
      //         cerr << index << " : ";
      //         for (size_t k = 0; k < s.size(); k++)
      //           cerr << s[k] << " ";
      //         cerr << "/ " << strings[j].second << endl;
      //       }
      double value = log_domain ? strings[j].second.Value()
          : exp(-strings[j].second.Value());
      //       if (normalize)
      //         value /= norm;
      features.push_back(make_pair(index, value));
      if (normalize) norm += value*value;
      // Sorted output instead
      //cout << index << ":" << value << " ";
    }
    if (normalize)
      norm = sqrt(norm);
    else
      norm = 1.0F;
    sort(features.begin(), features.end(), Compare<double>());
    for(ssize_t i = 0; i < features.size(); ++i)
      cout << features[i].first << ":"
           << features[i].second/norm << " ";
    cout << endl;
    delete fst;
  }


  if (FLAGS_feature_fst.size() > 0) {
    VectorFst<Arc> feature_fst;
    typename Arc::StateId start = feature_fst.AddState();
    typename Arc::StateId final = feature_fst.AddState();
    feature_fst.SetStart(start);
    feature_fst.SetFinal(final, Weight::One());
    for (ssize_t i = 0; i < string_vector.size(); ++i)
      for (ssize_t j = 0; j < string_vector[i].size(); ++j) {
        Arc arc;
        arc.ilabel = string_vector[i][j];
        arc.olabel = j ? 0 : i + 1;
        arc.weight = Weight::One();
        arc.nextstate = j == string_vector[i].size() - 1 ? final :
            feature_fst.AddState();
        feature_fst.AddArc(j ? feature_fst.NumStates() - 1 : start, arc);
      }
    feature_fst.Write(FLAGS_feature_fst);
  }
}


int main(int argc, char** argv) {
  string usage = "Compute the kernel matrix.\n\n  Usage: ";
  usage += argv[0];
  usage += " kernel.kar\n";

  std::set_new_handler(FailedNewHandler);
  SetFlags(usage.c_str(), &argc, &argv, true);
  if (argc > 2) {
    ShowUsage();
    return 1;
  }
  string in = argc > 1 ? string(argv[1]) : string();

  RationalKernel<LogArc> *kernel = RationalKernel<LogArc>::Read(in);
  if(!kernel)
    LOG(FATAL) << "klinfo: failed to read" << in;

  ssize_t n =
      (FLAGS_data_size == -1 || FLAGS_data_size > kernel->Data()->Size())
      ? kernel->Data()->Size()
      : FLAGS_data_size;

  ComputeFeatures(kernel, 0, n - 1, FLAGS_normalize, FLAGS_log_domain);

  delete kernel;
  return 0;
}
