
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#include <iostream>

#include "kernel/lib/learn-correlation.h"
#include "kernel/lib/kernel-learning-linearl2.h"
#include "kernel/lib/kernel-learning-polyl2.h"
#include "kernel/lib/io-utilities.h"

DEFINE_bool(sparse, false, "Use sparse matrix storage and operations.");
DEFINE_string(num_train, "0:0",
        "Use the points in the range N:M (i.e. N, N+1, ..., M for training.\n"
        "0:0 indicated using all the points");
DEFINE_string(lk_alg, "unif",
             "\tunif -- uniform weights (i.e. standard linear kernel)\n"
             "\tcorr -- correlation weights\n"
             "\tquadl2 -- l2-regularized polynomial combination\n"
             "\tlin2 -- l2-regularized linear combination");
DEFINE_double(corr_l1, 0, 
            "If corr_l1 > 0, regularize corr weights ||mu||_1 = corr_l1.");
DEFINE_double(corr_l2, 0, 
            "If corr_l2 > 0, regularize corr weights ||mu||^2_2 = corr_l2.");
DEFINE_double(alg_reg, 1.0,
              "Regularization paremeter used for learning algorithm\n"
              "Used only with the lin1 and lin2 method");
DEFINE_double(ker_reg, 1.0,
              "Regularization parameter used for the kernel selection.");
DEFINE_double(interp_param, 0.5,
              "Controls step size for gradient descent method used with\n"
              "lin2 method");
DEFINE_double(tol, 1e-10,
              "Stop criteria for gradient descent method used with\n"
              "lin2 method");
DEFINE_int32(max_iter, 50,
             "Maximum number of itereation used for lin2 gradient descent");
DEFINE_double(mu0, 0,
              "Set value of mu0 vector, for use with lin2 method");
DEFINE_bool(features, false,
            "Print out the feature vectors in libSVM format, instead of\n"
            "the kernel matrix.");
DEFINE_bool(center_labels, false,
            "Center the labels according to training data.");
DEFINE_bool(normalize_features, false,
            "Normalize feature to have norm one.");
DEFINE_bool(center_features, false,
            "Center features to have mean zero.");
DEFINE_double(offset, 0,
              "If offset > 0, include a constant feature equal to offset");
DEFINE_string(mu_file, "",
              "Save weights to output file.");
DEFINE_int32(approx, -1,
           "Use rank-approx svd decomposition for lin2 method. (EXPERIMENTAL)");
DEFINE_string(pred_file, "",
              "Output predictions to pred_file, if possible.");
//TODO(rostami): handle these flags appropriately.
DEFINE_bool(kar, false,
            "Save kernel as a OpenKernel kar object.");

using namespace kernel;


template <class M>
int klcorr(char** argv, M &data,
           DenseVector<typename M::ElementType> &labels) { //change data, labels to pointers
  if(data.NumRow() != labels.Size()) {
    LOG(ERROR) << "ERROR: #labels != #datapoints, formatting error?" << endl;
    LOG(ERROR) << "#labels: " << labels.Size()
         << " #datapoints: " << data.NumRow() << endl;
    return -1;
  }

  typedef typename M::ElementType ET;
  DenseMatrix<ET> kernel;

  LOG(INFO) << "Loaded " << data.NumRow() << " datapoints.";

  //SELECT TRAINING EXAMPLES
  M data_train;
  DenseVector<ET> labels_train;
  
  // parse num_train flag
  vector<char*> tokens;
  size_t train_start, train_end;
  char *cstr = new char[FLAGS_num_train.size()+1];
  strcpy(cstr, FLAGS_num_train.c_str());
  SplitToVector(cstr, ":", &tokens, true);
  if (tokens.size() != 2)
    LOG(FATAL) << "Format error in num_train flag";
  train_start = atoi(tokens[0]);
  train_end = atoi(tokens[1]);
  delete[] cstr;

  if (train_start < train_end) {
    DenseVector<size_t> rows;
    //MakeRange(train_start, train_end, &rows);
    MakeRangeWrap(train_start, train_end, data.NumRow(), &rows);
    LOG(INFO) << "Selecting " << rows.Size()
              << " training examples...";
    if (FLAGS_v > 2)
      PrintVector(rows, LOG(INFO));
    SubMatrixCopy(data, rows, &data_train);
    //SubsetCopy(train_start, train_end, labels, &labels_train);
    SubsetCopy(labels, rows, &labels_train);
  } else {
    data_train = data;
    labels_train = labels;
  }

  if (FLAGS_center_labels) {
    Sum(-1 * Mean(labels_train), DenseVector<double>(labels.Size(), 1.0),
        &labels);
    Sum(-1 * Mean(labels_train), DenseVector<double>(labels_train.Size(), 1.0),
        &labels_train);
  }

  if (FLAGS_center_features) {
    DenseVector<double> means;
    ComputeSums(data_train, &means);
    ScalarProduct(1.0 / data_train.NumRow(), &means);

    for (size_t j = 0; j < data.NumColumn(); ++j) {

      for (size_t i = 0; i < data.NumRow(); ++i)
        data.SetElement(i, j, data.Element(i, j) - means.Element(j));

      for (size_t i = 0; i < data_train.NumRow(); ++i)
        data_train.SetElement(i, j, 
             data_train.Element(i, j) - means.Element(j));

    }

//    PrintDenseMatrix(data, LOG(INFO));
  }

  if (FLAGS_normalize_features) {
    M temp;
    DenseVector<double> norms;
    ComputeSquaredNorms(data_train, &norms);
    for (size_t i = 0; i < norms.Size(); ++i)
      if (norms.Element(i) != 0)
        norms.SetElement(i, 1.0 / sqrt(norms.Element(i)));
    DiagonalProduct(data_train, norms, &temp);
    data_train = temp;
    DiagonalProduct(data, norms, &temp);
    data = temp;
  }

  //CALCULATE KERNEL MATRIX
  //note: the kernel uses only data_train to generate and
  //normalize corr-weights.
  DenseVector<ET> weights(data_train.NumColumn(), 0);
  DenseVector<ET> alg_weights;
  if (FLAGS_lk_alg.compare("unif") == 0) {
    DenseVector<ET> uniform(data_train.NumColumn(), 1);
    // Note, this normalize function squares the weights, but there is no
    // bad effect since they are uniform anyway.
    NormalizeCorrelationWeights(data_train, uniform, FLAGS_ker_reg, &weights);

  } else if (FLAGS_lk_alg.compare("corr") == 0) {
    MakeCorrelationWeights(data_train, labels_train, FLAGS_ker_reg,
		                       (FLAGS_offset > 0), &weights);

    if (FLAGS_corr_l1 > 0) {
      ScalarProduct(FLAGS_corr_l1 / Sum(weights), &weights);
    }
    if (FLAGS_corr_l2 > 0) {
      ScalarProduct(sqrt(FLAGS_corr_l2) / Norm(weights), &weights);
    }

  } else if (FLAGS_lk_alg.compare("lin2") == 0) {
    //TODO(rostami): treat offset feature specially here?
    DenseVector<ET> mu0(data_train.NumColumn(), 1);
    ScalarProduct(static_cast<ET>(FLAGS_mu0), &mu0);
    MakeLinearL2Weights(data_train, labels_train, mu0, FLAGS_ker_reg,
                        FLAGS_alg_reg, FLAGS_interp_param, FLAGS_tol,
                        FLAGS_max_iter, FLAGS_approx, &weights, &alg_weights);

  } else if (FLAGS_lk_alg.compare("quadl2") == 0) {
    vector< DenseMatrix<ET> > empty_vec;
    MakePolyL2Weights(empty_vec, data_train, labels_train, FLAGS_ker_reg,
                      FLAGS_alg_reg, FLAGS_interp_param, FLAGS_tol,
		      FLAGS_max_iter, &weights, &alg_weights);

  } else {
    LOG(ERROR) << "Unknown weight type" << endl;
    return 1;
  }
  LOG(INFO) << "Using " << weights.Size() << " features.";

  if (FLAGS_mu_file.compare("") != 0) {
    ofstream out(FLAGS_mu_file.c_str());
    if(!out.is_open()) {
      LOG(ERROR) << "Unable to save weights" << endl;
    } else {
      PrintVector(weights, out);
      out.close();
    }
  }

  ofstream outfile(argv[2]);
  if(!outfile.is_open()) {
      LOG(ERROR) << "Unable to save alg_weights" << endl;
      return 1;
  }

  if (FLAGS_features) {  //no need to compute kernel, print features
    if (FLAGS_lk_alg.compare("quadl2") == 0) {
      LOG(WARNING) << "explicit quadratic features not availible,"
                   << "printing kernel matrix instead.";

    } else {
      M features;
      MakeCorrelationFeatures(data, weights, &features);
      FeaturePrintLibSVM(features, labels, outfile);
      outfile.close();

      
      // Save predictions if requested and if availible
      if (FLAGS_pred_file.compare("") != 0 && alg_weights.Size() > 0) {
        DenseVector<double> pred;
        ofstream out(FLAGS_pred_file.c_str());
        if(!out.is_open()) {
          LOG(ERROR) << "Unable to save weights" << endl;
        } else {
          DenseVector<double> w;
          M f_train;
          typename M::TransposeMatrix f_train_t;
          if (train_start < train_end) {
          DenseVector<size_t> train_points;
          MakeRangeWrap(train_start, train_end,
                        data.NumRow(), &train_points);
          SubMatrixCopy(features, train_points, &f_train);
        } else {
          f_train = features;
        }
          f_train.Transpose(&f_train_t);
          Product(f_train_t, alg_weights, &w);
          Product(features, w, &pred);
          PrintVector(pred, out);
        }
        out.close();
      }
  
      return 0;
    }
  }

  MakeCorrelationKernel(data, data, weights, &kernel);
  if (FLAGS_lk_alg.compare("quadl2") == 0) {
    PointwisePower(2, &kernel);
  }
  KernelPrintLibSVM(kernel, labels, outfile);
  outfile.close();

  // Save predictions if requested and if availible
  if (FLAGS_pred_file.compare("") != 0 && alg_weights.Size() > 0) {
    DenseMatrix<double> k_test;
    if (train_start < train_end) {
      DenseVector<size_t> train_points;
      MakeRangeWrap(train_start, train_end,
                    data.NumRow(), &train_points);
      SubMatrixCopy(kernel, train_points, &k_test);
      k_test.Transpose(&k_test);
    } else {
      k_test = kernel;
    }

    ofstream out(FLAGS_pred_file.c_str());
    if(!out.is_open()) {
      LOG(ERROR) << "Unable to save weights" << endl;
    } else {
      DenseVector<double> pred;
      KRRDualPredict(k_test, alg_weights, &pred);

      PrintVector(pred, out);
    }
  }

  return 0;
}


int main(int argc, char** argv) {
  string usage =
      "Print precomputed correlation based kernel (libSVM format),\n"
      "input is expected to be in libSVM format.\n\n";
  usage += "Usage: " + string(argv[0]);
  usage += " data outfile\n";

  std::set_new_handler(FailedNewHandler);
  SetFlags(usage.c_str(), &argc, &argv, true);

  if (argc < 3) {
    ShowUsage();
    return 1;
  }

  DenseVector<double> labels;
  if (FLAGS_sparse) {
    SemiDenseMatrix<SparseVector<double>, DENSE_BY_ROW> data;
    if(LibSVMToSparseMatrix(argv[1], FLAGS_offset, &data, &labels) != 0) {
      LOG(ERROR) << "Error reading input file" << endl;
      return -1;
    }
    return klcorr(argv, data, labels);
  } else {
    DenseMatrix<double> data;
    if(LibSVMToDenseMatrix(argv[1], FLAGS_offset, &data, &labels) != 0) {
      LOG(ERROR) << "Error reading input file" << endl;
      return -1;
    }
    return klcorr(argv, data, labels);
  }
}
