
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#include "kernel/lib/io-utilities.h"
#include "kernel/lib/krr.h"

DEFINE_int32(cross_validate, 0,
             "Do n-fold cross-validation and return RMSE\n"
             "NOTE: does not shuffle data.");
// TODO(rostami): implement this, centering can be done straightforwardly if
// we have explicit feature vectors, or done using the kernel function:
//   K_centered(x_i,x_j)
//      = K(x_i,x_j) - K(x_i, x_bar) - K(x_j, x_bar) + K(x_bar, x_bar).
DEFINE_bool(center, false, "Account for the lack of an offset by centering\n"
            "the labels and data (in the feature space) [TODO]");
DEFINE_bool(kernel, false, "Data argument is a precomputed kernel in\n"
            "libSVM format.");
DEFINE_bool(kar, false, "Data argument is an OpenKernel kar object. [TODO]");
DEFINE_bool(sparse, false, "Use sparse feature representation for data.");
DEFINE_bool(dual, false, "Force the use of the dual problem.");
DEFINE_bool(primal, false,
            "Force the use of the primal problem (when possible).");
DEFINE_int32(approx, 0,
             "Find an approximate solution using the indicated lower rank,\n"
             "for use only with feature vector imput.");

using namespace kernel;

double RMSE(const DenseVector<double> &v1,
            const DenseVector<double> &v2) {
  CHECK(v1.Size() == v2.Size());
  double rmse = 0;
  for (size_t i = 0; i < v1.Size(); ++i)
    rmse += pow((v1.Element(i) - v2.Element(i)), 2);

  rmse = sqrt(rmse / v1.Size());
  return rmse;
}

//sort in ascending order on first element
bool compare_pair(const pair<size_t, double> &a,
                  const pair<size_t, double> &b) {
  if(a.first < b.first)
    return true;
  else
    return false;
}

//TODO(rostami): read/write data in binary format.
void KRRPrintModel(const DenseVector<size_t> &ids,
                   const DenseVector<double> &vars,
                   const string &outfile,
                   const bool is_dual,
                   const bool is_centered) {
  CHECK(ids.Size() == vars.Size());
  ofstream out(outfile.c_str());
  CHECK(out.is_open());

  list< pair<size_t, double> > pairs;
  for(size_t i = 0; i < ids.Size(); ++i) {
    pair<size_t, double> temp(ids.Element(i), vars.Element(i));
    pairs.push_back(temp);
  }
  pairs.sort(compare_pair);

  if (is_dual)
    out << "dual_vars" << endl;
  else
    out << "primal_vars" << endl;
  if (is_centered)
    out << "centered" << endl;
  else
    out << "not_centered" << endl;

  for(list< pair<size_t, double> >::iterator it = pairs.begin();
      it != pairs.end(); ++it)
    out << it->first << " " << it->second << endl;

  out.close();
}

template <class M>
int KRRTrainMain(const M &data, DenseMatrix<typename M::ElementType> &kernel,
                 const DenseVector<typename M::ElementType> &labels,
                 double reg_param, const DenseVector<size_t> &ids,
                 const string &outfile) {

  int size = labels.Size();

  //CROSS-VALIDATION ROUTINE
  if (FLAGS_cross_validate > 0) {
    int n = FLAGS_cross_validate;
    int step = size / n;
    int start, stop;
    int num_train = size - step;
    int num_test = step;

    DenseMatrix<double> train_kernel(num_train, num_train, 0);
    DenseMatrix<double> test_kernel(num_test, num_train, 0);

    DenseVector<double> train_labels(num_train, 0);
    DenseVector<double> test_labels(num_test, 0);

    DenseVector<int> train_index(num_train, 0);
    DenseVector<int> test_index(num_test, 0);

    DenseVector<double> dual_vars(train_index.Size(), 0);
    DenseVector<double> pred(test_index.Size(), 0);

    DenseVector<double> rmse_vec(n, 0);

    for (size_t i = 0; i < n; ++i) {
      LOG(INFO) << "FOLD:" << i;
      start = i * step;
      stop = start + step;  //[start,stop] defines the interval of test points
      for (size_t j = 0; j < size; ++j) {
        if (j < start) {
          train_index.SetElement(j, j);
          train_labels.SetElement(j, labels.Element(j));
        } else if (j < stop) {
          test_index.SetElement(j - start, j);
          test_labels.SetElement(j - start, labels.Element(j));
        }
        else {
          train_index.SetElement(j - step, j);
          train_labels.SetElement(j - step, labels.Element(j));
        }
      }

      if (FLAGS_kar) {
        //TODO(rostami): implement these functions in lib/krr.h
        /*
        if (dual_is_better) {
          KRRDualTrain(kernel_object, train_index, train_labels, reg_param
                      &dual_vars);
          KRRDualPredict(kernel_object, train_index, test_index, dual_vars,
                         &pred);
        } else {
          KRRPrimalTrain(kernel_object, train_index, train_labels, reg_param
                       &primal_vars);
          KRRDualPredict(kernel_object, train_index, test_index, primal_vars,
                         &pred);
        */
      } else {
        SubMatrixCopy(kernel, train_index, train_index, &train_kernel);
        SubMatrixCopy(kernel, test_index, train_index, &test_kernel);

        KRRDualTrain(train_kernel, train_labels, reg_param, &dual_vars);
        KRRDualPredict(test_kernel, dual_vars, &pred);
      }
      double rmse = RMSE(pred, test_labels);

      DenseVector<double> mean_label(pred.Size(), Mean(train_labels));
      double sanity_rmse = RMSE(mean_label, test_labels);

      rmse_vec.SetElement(i, rmse);
      LOG(INFO) << "RMSE: " << rmse;
      if (sanity_rmse < rmse) {
        LOG(WARNING)
            << "WARNING: prediction does not outperform guessing the mean.";
        LOG(INFO) << "Sanity RMSE: " << sanity_rmse;
      }
    }
    LOG(INFO) << "Avg RMSE: " << Mean(rmse_vec)
         << "Std RMSE: " << StdDev(rmse_vec);

    return 0;
  } // end CROSS-VALIDATION ROUTINE

  //NORMAL TRAINING
  DenseVector<double> dual_vars;
  DenseVector<double> primal_vars;
  if (FLAGS_kar) {
    //decide to solve in primal or dual
  } else if (FLAGS_kernel) {
    //must solve in dual
    if (FLAGS_approx > 0) { //approximate dual
      if (FLAGS_approx > data.NumRow())
        LOG(ERROR) << "Error: Approximate rank must be less than"
                   "the number of instances." << endl;
      DenseVector<size_t> sample_index;
      LOG(ERROR) << "Approximate method not yet implemented for kernel input"
                 << endl;
      // TODO(rostami): If approx > 0, should only read subset of kernel
      // from file i.e. kernel = C in that case
      // KRRApproxDualTrain(C, labels, reg_param, &dual_vars);
    } else {
      KRRDualTrain(kernel, labels, reg_param, &dual_vars);
    }
    KRRPrintModel(ids, dual_vars, outfile, true, FLAGS_center);
  } else {
    size_t m = data.NumRow();
    size_t n = data.NumColumn();
    // TODO(Afshin): pow(x, y) x must be a double or float
    // there is no pow(int, int) because of overflows.
    // So use cast to double or n*n*n.
    //     size_t pcomplexity = pow(n, 3) + pow(n, 2) + m * n;
    //     size_t dcomplexity = pow(m, 3) + pow(m, 2);
    double pcomplexity = pow(static_cast<double>(n), 3) +
        pow(static_cast<double>(n), 2) + m * n;
    double dcomplexity = pow(static_cast<double>(m), 3) +
        pow(static_cast<double>(m), 2);

    if(FLAGS_dual || (!FLAGS_primal && dcomplexity < pcomplexity)) {
      LOG(INFO) << "Solving in dual.";
      typename M::TransposeMatrix data_t;
      data.Transpose(&data_t);

      if (FLAGS_approx > 0) { //approximate dual
        if (FLAGS_approx > data.NumRow())
          LOG(ERROR) << "Error: Approximate rank must be less than"
                     "the number of instances" << endl;
        DenseVector<size_t> sample_index;
        MakeRange(0, FLAGS_approx, &sample_index);
        KRRApproxDualTrain(data, sample_index, labels, reg_param, &dual_vars);

      } else { // full dual
        Product(data, data_t, &kernel);
        KRRDualTrain(kernel, labels, reg_param, &dual_vars);
      }
      Product(data_t, dual_vars, &primal_vars);

    } else {
      LOG(INFO) << "Solving in primal.";
      if (FLAGS_approx > 0)
        LOG(ERROR)
          << "Approximate method not yet implemented for primal solver" << endl;
      KRRPrimalTrain(data, labels, reg_param, &primal_vars);
    }
    DenseVector<size_t> feature_ids;
    MakeRange(0, data.NumColumn(), &feature_ids);
    KRRPrintModel(feature_ids, primal_vars, outfile, false, FLAGS_center);
  }

  return 0;
}

int main(int argc, char** argv) {
  string usage =
      "Train KRR model using a precomputed kernel.\n\n";
  usage += "Usage: " + string(argv[0]);
  usage += " train_data krr_reg [outfile]\n";

  std::set_new_handler(FailedNewHandler);
  SetFlags(usage.c_str(), &argc, &argv, true);

  if (argc < 3) {
    ShowUsage();
    return 1;
  }

  string outfile;
  if(argc >= 4) {
    outfile = argv[3];
  } else {
    outfile = argv[1];
    outfile += ".model";
  }

  double reg_param = strtod(argv[2], NULL);
  DenseVector<double> labels;
  DenseVector<size_t> ids;
  DenseMatrix<double> kernel, data_dense;
  SemiDenseMatrix<SparseVector<double>, DENSE_BY_ROW> data_sparse;

  if (FLAGS_kar) {
    //read kernel object using binary Read function

  } else if (FLAGS_kernel) {
    DenseMatrix<double> kernel;
    if (LibSVMKernelToDense(string(argv[1]), &kernel, &labels, &ids) != 0) {
      LOG(ERROR) << "Error: problem reading kernel file." << endl;
      return -1;
    }
    return KRRTrainMain(data_dense, kernel, labels, reg_param, ids, outfile);

  } else {
    if (FLAGS_sparse) {
      SemiDenseMatrix<SparseVector<double>, DENSE_BY_ROW> data;
      if (LibSVMToSparseMatrix(argv[1], &data_sparse, &labels) != 0) {
        LOG(ERROR) << "Error: problem reading sparse data file." << endl;
        return -1;
      }
      return KRRTrainMain(data_sparse, kernel, labels, reg_param, ids, outfile);
    } else {
      DenseMatrix<double> data;
      if (LibSVMToDenseMatrix(argv[1], &data_dense, &labels) != 0) {
        LOG(ERROR) << "Error: problem reading dense data file." << endl;
        return -1;
      }
      return KRRTrainMain(data_dense, kernel, labels, reg_param, ids, outfile);
    }
  }
}
