
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#ifndef KERNEL_LIB____LIB_IO_UTILITIES_H_
#define KERNEL_LIB____LIB_IO_UTILITIES_H_

#include <vector>
#include <cstdlib>

#include "kernel/lib/dense-matrix.h"
#include "kernel/lib/semidense-matrix.h"
#include "kernel/lib/dense-vector.h"
#include "kernel/lib/sparse-vector.h"
#include "kernel/lib/kernel.h"

namespace kernel {

#define ERR_BADFORMAT 1
#define ERR_OUTOFBOUNDS 2
#define ERR_FILENOTFOUND 3

typedef vector<vector<double> > Matrix;
typedef vector<double> Vector;

void AsciiToVector(const string &filename, Vector *vec);

void AsciiToMatrix(const string &filename, Matrix *mat);

// Expects each line to be of the form index_1 index_2 val
// (r,c) is the dimension of the data.
// Outputs a DENSE matrix.
int SparseAsciiToMatrix(const string &filename, int r, int c, Matrix *mat);

void KernelPrintLibSVM(const Matrix &mat, const Vector &labels);

template <typename T>
void AsciiToDenseVector(const string &filename, DenseVector<T> *vec);

template <typename T>
void MatrixToDenseMatrix(const Matrix &m, DenseMatrix<T> *dm);

template <typename T>
void AsciiToDenseMatrix(const string &filename, DenseMatrix<T> *mat);

template <typename T>
int SparseAsciiToDenseMatrix(const string &filename, int r, int c,
                             DenseMatrix<T> *mat);

template <typename T>
int LibSVMToDenseMatrix(const string &filename,
                             DenseMatrix<T> *mat,
                             DenseVector<T> *labels);

template <typename T, typename S>
int LibSVMToSparseMatrix(const string &filename,
                         S *mat,
                         DenseVector<T> *labels);

template <typename T>
void KernelPrintLibSVM(const DenseMatrix<T> &mat,
                       const DenseVector<T> &labels,
                       ostream &out_stream);

template <typename T>
void KernelPrintLibSVM(const DenseMatrix<T> &mat,
                       const DenseVector<T> &labels,
											 const DenseVector<T> &ids,
                       ostream &out_stream);

template <typename T> //deprecated
int LibSVMKernelToDense(const string &filename,
                         int size,
                         DenseMatrix<T> *mat,
                         DenseVector<T> *labels);

template <typename T, typename I>
int LibSVMKernelToDense(const string &filename,
                         DenseMatrix<T> *mat,
                         DenseVector<T> *labels,
                         DenseVector<I> *point_ids);

template <typename T>
int KarToDense(const string &filename, DenseMatrix<T> *mat);

template <class V> void PrintVector(const V &vec, ostream &out);

template <typename T> void PrintDenseMatrix(const DenseMatrix<T> &mat, 
                                            ostream &out);

template <typename T>
void AsciiToDenseVector(const string &filename, DenseVector<T> *vec) {
  string line;
  ifstream file(filename.c_str());
  CHECK(file.is_open());
  uint64 i = 0;
  while(1) {
    getline(file, line);
    if(file.eof()) break;

    vec->AddElement(i, strtod(line.c_str(), NULL));
    i++;
  }
  file.close();
}

template <typename T>
void MatrixToDenseMatrix(const Matrix &m, DenseMatrix<T> *dm) {
  size_t nr = m.size();
  size_t nc = m[0].size();
  dm->Reset(nr, nc, 0.0);

  for(size_t i = 0; i < nr; ++i)
    for(size_t j = 0; j < nc; ++j)
      dm->SetElement(i, j, m[i][j]);
}

template <typename T>
void AsciiToDenseMatrix(const string &filename, DenseMatrix<T> *dm) {
  Matrix m = Matrix();

  AsciiToMatrix(filename, &m);
  MatrixToDenseMatrix(m, dm);
}

template <typename T>
int SparseAsciiToDenseMatrix(const string &filename, int r, int c,
                             DenseMatrix<T> *mat) {
  string line;
  char *index, *oindex;
  ifstream file(filename.c_str());
  CHECK(file.is_open());
  double temp[3];

  if(mat->NumRow() != r || mat->NumColumn() != c)
    mat->Reset(r, c, 0);

  while(1) {
    getline(file, line);
    if(file.eof()) break;

    temp[0] = strtod(line.c_str(), &index);
    oindex = index;
    temp[1] = strtod(oindex, &index);
    oindex = index;
    temp[2] = strtod(oindex, &index);
    if(oindex == index) return ERR_BADFORMAT;

    if(temp[0] >= r || temp[1] >= c) return ERR_OUTOFBOUNDS;
    mat->SetElement(static_cast<int>(temp[0]),
                    static_cast<int>(temp[1]),
                    temp[2]);
  }
  file.close();
  return 0;
}

template <typename T>
int SparseAsciiToDenseMatrix(const string &filename, DenseMatrix<T> *mat) {
  string line;
  char *index, *oindex;
  int max_r = 0;
  int max_c = 0;
  ifstream file(filename.c_str());
  CHECK(file.is_open());
  int row, col;

  //count maximum number of points and features
  while(1) {
    getline(file, line);
    if(file.eof()) break;

    row = static_cast<int>(strtod(line.c_str(), &index));
    oindex = index;
    col = static_cast<int>(strtod(oindex, &index));

    if(row > max_r)
      max_r = row;
    if(col > max_c)
      max_c = col;
  }
  file.close();

  //do a second pass, now that size of matrix is determined.
  return SparseAsciiToDenseMatrix(filename, max_r + 1, max_c + 1, mat);
}

template<typename T>
int LibSVMGetLabels(const string &filename,
                    DenseVector<T> *labels,
                    uint64 *max_feature) {
  string line;
  char *cstr;
  vector<char*> tokens, subtokens;
  uint64 i, id;
  *max_feature = 0;

  ifstream file(filename.c_str());
  CHECK(file.is_open());
  labels->Resize(1, 0);
  i = 0;
  while (1) {
    getline(file, line);
    if (file.eof())
      break;
    cstr = new char[line.size()+1];
    strcpy(cstr, line.c_str());

    tokens.clear();
    SplitToVector(cstr, " \t", &tokens, true);
    if (tokens.size() < 1) {
      delete[] cstr;
      LOG(ERROR) << "Error -- line: " << i << " missing label" << endl;
      return ERR_BADFORMAT;
    }
    labels->AddElement(i, strtod(tokens[0], NULL));
    ++i;
    if (tokens.size() < 2) { // no data
      delete[] cstr;
      continue;
    }

    subtokens.clear();
    SplitToVector(tokens[tokens.size() - 1], ":", &subtokens, true);
    if ((subtokens.size() != 2) || (atol(subtokens[0]) <= 0)) {
      delete[] cstr;
      LOG(ERROR) << "Error -- line: " << i
                << " subtokens.size(): " << subtokens.size()
                << " id_val: " << atol(subtokens[0]) << endl;
      return ERR_BADFORMAT;
    }

    id = atol(subtokens[0]);
    if (id == 0) {
      delete[] cstr;
      return ERR_BADFORMAT;
    }
    if (*max_feature < id)
      *max_feature = id;

    delete[] cstr;
  }
  file.close();

  return 0;
}

// Read a libSVM formatted data file and populate a dense data matrix
// and label vector.
//
// Each line is of the form:
// <label> <index1>:<value1> <index2>:<value2> ...
//
// indices must be listed in ascending order.
template <typename T>
int LibSVMToDenseMatrix(const string &filename,
                             DenseMatrix<T> *mat,
                             DenseVector<T> *labels) {
  return LibSVMToDenseMatrix(filename, static_cast<T>(0), mat, labels);
}

template <typename T>
int LibSVMToDenseMatrix(const string &filename,
                        T offset,
                        DenseMatrix<T> *mat,
                        DenseVector<T> *labels) {
  string line;
  char *cstr;
  vector<char*> tokens, subtokens;
  uint64 i, id, max_feature = 0;
  T value;
  int flag;

  if((flag = LibSVMGetLabels(filename, labels, &max_feature)) != 0)
    return flag;

  ifstream file(filename.c_str());
  CHECK(file.is_open());
  uint64 num_points = labels->Size();
  uint64 num_features = max_feature;
  if (offset > 0)
    ++num_features;
  mat->Reset(num_points, num_features, 0);
  i = 0;
  while (1) {
    getline(file, line);
    if (file.eof())
      break;
    cstr = new char[line.size()+1];
    strcpy(cstr, line.c_str());

    tokens.clear();
    SplitToVector(cstr, " \t", &tokens, true);
    if (tokens.size() < 2) { //no data
      delete[] cstr;
      ++i;
      continue;
    }

    for (size_t j = 1; j < tokens.size(); ++j) {
      subtokens.clear();
      SplitToVector(tokens[j], ":", &subtokens, true);
      id = atol(subtokens[0]);
      value = strtod(subtokens[1], NULL);
      if (id > num_features) {
        delete[] cstr;
        return ERR_OUTOFBOUNDS;
      }
      mat->SetElement(i, id - 1, value);
    }

    delete[] cstr;
    ++i;
  }

  if (offset > 0)
    for (size_t i = 0; i < mat->NumRow(); ++i)
      mat->SetElement(i, num_features - 1, offset);

  return 0;
}

template <typename T, typename S>
int LibSVMToSparseMatrix(const string &filename,
                         S *mat,
                         DenseVector<T> *labels) {
  return LibSVMToSparseMatrix(filename, static_cast<T>(0), mat, labels);
}

template <typename T, typename S>
int LibSVMToSparseMatrix(const string &filename,
                         T offset,
                         S *mat,
                         DenseVector<T> *labels) {
  string line;
  char *cstr;
  vector<char*> tokens, subtokens;
  uint64 id, max_feature;
  T value;
  int flag;

  if((flag = LibSVMGetLabels(filename, labels, &max_feature)) != 0)
    return flag;

  ifstream file(filename.c_str());
  CHECK(file.is_open());
  while (1) {
    getline(file, line);
    if (file.eof())
      break;
    cstr = new char[line.size()+1];
    strcpy(cstr, line.c_str());

    tokens.clear();
    SplitToVector(cstr, " \t", &tokens, true);

    SparseVector<T> row;
    if (tokens.size() < 2) { //no data
      mat->AddRow(row);
      delete[] cstr;
      continue;
    }
    for (size_t j = 1; j < tokens.size(); ++j) {
      subtokens.clear();
      SplitToVector(tokens[j], ":", &subtokens, true);
      id = atol(subtokens[0]);
      value = strtod(subtokens[1], NULL);
      if (value == 0)
        continue;
      row.AddElement(id - 1, value);
    }
    if (offset > 0)
      row.AddElement(max_feature, offset);
    mat->AddRow(row);

    delete[] cstr;
  }

  return 0;
}

template <typename T>
void KernelPrintLibSVM(const DenseMatrix<T> &mat,
                       const DenseVector<T> &labels,
                       ostream &out_stream) {
  CHECK(labels.Size() == mat.NumRow());
  int i,j;
  for(i = 0; i < mat.NumRow(); ++i) {
    out_stream << labels.Element(i) << " 0:" << i+1 << " ";
    for(j = 0; j < mat.NumColumn(); ++j)
      out_stream << j + 1 << ":" << mat.Element(i, j) << " ";
    out_stream << endl;
  }
}

template <typename T>
void KernelPrintLibSVM(const DenseMatrix<T> &mat,
                       const DenseVector<T> &labels,
											 const DenseVector<T> &ids,
                       ostream &out_stream) {
  CHECK(labels.Size() == mat.NumRow());
  int i,j;
  for(i = 0; i < mat.NumRow(); ++i) {
    out_stream << labels.Element(i) << " 0:" << i+1 << " ";
    for(j = 0; j < mat.NumColumn(); ++j)
      out_stream << j + 1 << ":" << mat.Element(i, j) << " ";
    out_stream << endl;
  }
}

template <typename T>
void FeaturePrintLibSVM(const SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW>
                        &mat,
                        const DenseVector<T> &labels,
                        ostream &out_stream) {
  CHECK(labels.Size() == mat.NumRow());
  SparseVector<T> row;
  for (size_t i = 0; i < mat.NumRow(); ++i) {
    row = mat.Row(i);
    out_stream << labels.Element(i) << " ";
    for (size_t j = 0; j < row.Size(); ++j)
      out_stream << row.Pair(j).first + 1 << ":" << row.Pair(j).second << " ";
    out_stream << endl;
  }
}

template <typename T>
void FeaturePrintLibSVM(const DenseMatrix<T> &mat,
                        const DenseVector<T> &labels,
                        ostream &out_stream) {
  CHECK(labels.Size() == mat.NumRow());
  for (size_t i = 0; i < mat.NumRow(); ++i) {
    out_stream << labels.Element(i) << " ";
    for (size_t j = 0; j < mat.NumColumn(); ++j) {
      if (mat.Element(i,j) != 0)
        out_stream << j + 1 << ":" << mat.Element(i,j) << " ";
    }
    out_stream << endl;
  }
}

// Simple libSVM format reader, assumes data is size-by-size, and
// that row and columns are ordered (i.e. ignores point and feature
// ids.
template <typename T>
int LibSVMKernelToDense(const string &filename,
                         int size,
                         DenseMatrix<T> *mat,
                         DenseVector<T> *labels) {

  if(mat->NumRow() != size || mat->NumColumn() != size)
    mat->Reset(size, size, 0);
  if(labels->Size() != size)
    labels->Resize(size, 0);

  string line;
  char *tok, *cstr;
  double label, value;
  int i, j, tmp;
  ifstream file(filename.c_str());
  CHECK(file.is_open());
  while(1) {
    getline(file, line);
    if(file.eof()) break;

    cstr = new char[line.size()+1];
    strcpy(cstr, line.c_str());

    //read label and index
    tok = strtok(cstr, " \t");
    label = strtod(tok, NULL);
    tok = strtok(NULL, ":");
    tmp = atoi(tok);
    tok = strtok(NULL, " \t");
    i = atoi(tok) - 1;

    if(i >= size || i < 0) return ERR_OUTOFBOUNDS;
    if(tmp != 0) return ERR_BADFORMAT;
    labels->SetElement(i, static_cast<T>(label));

    //read features
    while(1) {
      tok = strtok(NULL, ":");
      if (tok == NULL) break;
      j = atoi(tok) - 1;
      tok = strtok(NULL, " \t");
      if (tok == NULL) break;
      value = strtod(tok, NULL);

      if (j >= size || j < 0) return ERR_OUTOFBOUNDS;

      mat->SetElement(i, j, static_cast<T>(value));
    }

    delete[] cstr;
  }

  return 0;
}

// Reads libSVM format kernel file, allowing for permutations and a
// subset of the rows of the original file.  Will produce a m-by-m
// matrix, where m is the number of rows of the kernel file.
//
// Lines may be in any order, but the point ids on a line must be
// ordered, dense, and contiguous (i.e. per the libSVM format
// requirements).
//
// The returned DenseMatrix has the rows and columns sorted in the
// same order as the rows of the original text file.
template <typename T, typename I>
int LibSVMKernelToDense(const string &filename,
                         DenseMatrix<T> *mat,
                         DenseVector<T> *labels,
                         DenseVector<I> *point_ids) {
  string line;
  char *cstr;
  vector<char*> tokens, subtokens;
  uint64 i, id;
  T value;

  //FIRST PASS: Store point ids and see how many training points there are.
  ifstream file(filename.c_str());
  CHECK(file.is_open());
  labels->Resize(1, 0);
  point_ids->Resize(1, 0);
  i = 0;
  while (1) {
    getline(file, line);
    if (file.eof()) break;
    cstr = new char[line.size()+1];
    strcpy(cstr, line.c_str());

    tokens.clear();
    SplitToVector(cstr, " \t", &tokens, true);
    if (tokens.size() < 2) {
      delete[] cstr;
      return ERR_BADFORMAT;
    }

    subtokens.clear();
    SplitToVector(tokens[1], ":", &subtokens, true);
    if ((subtokens.size() != 2) || (atol(subtokens[0]) != 0)) {
      delete[] cstr;
      return ERR_BADFORMAT;
    }

    id = atol(subtokens[1]);
    if (id == 0) {
      delete[] cstr;
      return ERR_BADFORMAT;
    }

    labels->AddElement(i, strtod(tokens[0], NULL));
    point_ids->AddElement(i, id);

    delete[] cstr;

    ++i;
  }
  file.close();

  //SECOND PASS: Store kernel matrix.
  mat->Reset(labels->Size(), labels->Size(), 0);
  file.open(filename.c_str());
  CHECK(file.is_open());
  i = 0;
  while (1) {
    getline(file, line);
    if (file.eof()) break;
    cstr = new char[line.size()+1];
    strcpy(cstr, line.c_str());

    tokens.clear();
    SplitToVector(cstr, " \t", &tokens, true);
    for (size_t j = 0; j <  point_ids->Size(); ++j) {
      subtokens.clear();
      SplitToVector(tokens[static_cast<int>(point_ids->Element(j)) + 1], 
                           ":", &subtokens, true);
      //SplitToVector(tokens[point_ids->Element(j) + 1], ":", &subtokens, true);
      if (subtokens.size() != 2) {
       delete[] cstr;
       return ERR_BADFORMAT;
      }
      id = atol(subtokens[0]);
      value = strtod(subtokens[1], NULL);
      if ((i > mat->NumRow())
         || (id != point_ids->Element(j))
         || (j > mat->NumColumn())) {
        LOG(ERROR) << "ERROR: accessing index (" <<  i << "," << j
             << ") in size " << mat->NumRow() << " matrix." << endl;
        LOG(ERROR) << "Parsed point_id:" << id << " expected:"
             << point_ids->Element(j) << endl;
        delete[] cstr;
        return ERR_OUTOFBOUNDS;
      }
      mat->SetElement(i, j, value);
    }
    delete[] cstr;
    ++i;
  }

  return 0;
}

template <typename T>
int KarToDense(const string &filename, DenseMatrix<T> *mat) {
  KernelBase *kernel = KernelBase::Read(filename);

  if(!kernel) LOG(FATAL) << "unable to read kar file";

  int size = kernel->Data()->Size();
	mat->Reset(size, size, 0);

	for (int i = 0; i < size; ++i)
	  for (int j = 0; j < size; ++j)
		  mat->SetElement(i, j, kernel->Eval(i,j));
}

template <class V> void PrintVector(const V &vec, ostream &out) {
  for (size_t i = 0; i < vec.Size(); ++i)
    out << i << ":" << vec.Element(i) << endl;
}

template <typename T> void PrintDenseMatrix(const DenseMatrix<T> &mat, 
                                            ostream &out) {
  for (size_t i = 0; i < mat.NumRow(); ++i) {
    for (size_t j = 0; j < mat.NumColumn(); ++j)
      out << mat.Element(i,j) << "\t";
    out << endl;
    }
}

} //namespace kernel

#endif  // KERNEL_LIB____LIB_IO_UTILITIES_H_
