
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#ifndef KERNEL_LIB_KERNEL_LEARNING_POLYL2_H_
#define KERNEL_LIB_KERNEL_LEARNING_POLYL2_H_

#include <cmath>

#include "kernel/lib/dense-matrix.h"
#include "kernel/lib/dense-vector.h"
#include "kernel/lib/learn-correlation.h"
#include "kernel/lib/kernel-learning-linearl2.h"
#include "kernel/lib/krr.h"

namespace kernel {

//
// HELPER FUNCTIONS
//

// In-place pointwise exponentiation.
template <typename T> inline
void PointwisePower(double p, DenseMatrix<T> *mat) {
  for (size_t i = 0; i < mat->NumRow(); ++i)
    for (size_t j = 0; j < mat->NumColumn(); ++j)
      mat->SetElement(i, j, pow(mat->Element(i, j), p));
}

// Compute the gradient of quadratic combination objective.
// Also, returns value of objective function and the value
// of the KRR variables.
template <typename T> //inline
double QuadGradient(const vector< DenseMatrix<T> > &kernels,
                    const DenseVector<T> &weights,
                    const DenseVector<T> &labels,
                    double &reg_param,
                    DenseVector<T> *grad,
                    DenseVector<T> *alpha) {
  VLOG(2) << "general gradient";
  CHECK(weights.Size() == kernels.size());

  DenseMatrix<T> temp, temp_i;
  LinearCombination(kernels, weights, &temp);
  PointwisePower(2, &temp);

  DenseVector<T> temp_vec;
  //alpha = y*((sum_i w_i * K_i).^2 + lambda*I)^{-1}
  KRRDualTrain(temp, labels, reg_param, alpha);

  LinearCombination(kernels, weights, &temp); //note: recomputing in order
                                              //to conserve memory.
  //grad_j = -2 * alpha^T * ((sum_i w_i * K_i) \circ K_j) * alpha
  if (grad->Size() != weights.Size())
    grad->Resize(weights.Size(), 0);
  for (size_t i = 0; i < weights.Size(); ++i) {
    PointwiseProduct(temp, kernels[i], &temp_i);
    Product(temp_i, *alpha, &temp_vec);
    grad->SetElement(i, -2 * DotProduct(*alpha, temp_vec));
  }

  return DotProduct(*alpha, labels);
}

// Compute the gradient of quadratic combination objective, in the
// case where base kernels correspond to features.
// Also, returns value of objective function and the value
// of the KRR variables.
template <class M, typename T> //inline
double QuadGradientRank1(const M &data,
                    const DenseVector<T> &weights,
                    const DenseVector<T> &labels,
                    double &reg_param,
                    DenseVector<T> *grad,
                    DenseVector<T> *alpha) {
  VLOG(2) << "rk1 gradient";
  CHECK(weights.Size() == data.NumColumn());
  DenseMatrix<T> temp, temp_i, kernel_i;
  MakeCorrelationKernel(data, data, weights, &temp);
  PointwisePower(2, &temp);

  VLOG(2) << "compute alpha";
  DenseVector<T> temp_vec;
  // alpha = y*((sum_i w_i * K_i).^2 + lambda*I)^{-1}
  KRRDualTrain(temp, labels, reg_param, alpha);

  VLOG(2) << "compute linear kernel";
  MakeCorrelationKernel(data, data, weights, &temp);// note: recomputing in
                                                   // order to conserve memory.
  // grad_j = -2 * alpha^T * ((sum_i w_i * K_i) \circ K_j) * alpha
  if (grad->Size() != weights.Size())
    grad->Resize(weights.Size(), 0);

  vector< SparseVector<T> > columns;
  MatrixToColumnVectors(data, &columns);
  for (size_t i = 0; i < weights.Size(); ++i) {
    VLOG(3) << "compute partial derivative:" << i;
    // kernel_i = x_i x_i^top, where x_i is ith feature
    OuterProduct(columns[i], data.NumRow(),
                 columns[i], data.NumRow(), &kernel_i);
    PointwiseProduct(temp, kernel_i, &temp_i);
    Product(temp_i, *alpha, &temp_vec);
    grad->SetElement(i, -2 * DotProduct(*alpha, temp_vec));
  }

  return DotProduct(*alpha, labels);
}


//
// MAIN FUNCTIONS
//

// Regularize according to ||weights|| < ker_reg
// If kernels.size() == 1, assumes rk1 kernels based on features
//
// Learn a quadratic combination of basekernels which optimized the
// kernel ridge regression objective, parametrized by weights:
//   K = (\sum_i weight[i]  K_i)^2
//
// Where the base kernels are either provided in the vector kernels,
// or if kernels.size() == 0 each feature in the data matrix is used
// to construct a rank-1 kernel.
// The combination weights are regularized so that, ||weights|| < ker_reg.
//
// INPUT:
// kernels - a vector of base kernel matrices.
// data - a Dense or SemiDense matrix, which each row is a training instance
// (used only if kernels.size() == 0).
// labels - the corresponding vector of labels.
// kernel_reg - limit the norm of the generated weights.
// alg_reg - the regularization parameter that will be used by KRR problem.
// interp_param - a value between 0 and 1, values closer to 0 will
//   allow the algorithm to take bigger steps, leading to faster
//   convergence, but possibly also making the algorithm instable. A
//   good initial choice is 0.5.
// stop_tol - the stopping criteria uses this tolerance to decide when
//   to stop, a smaller value will give a more accurate solution, but
//   take longer to converge.
// max_iter - maximum number of iterations, will force function to
//   return even if stop_tol is not met.
//
// OUTPUT:
// weights - the kernel (feature) weighting (mu).
// krr_weights - the dual variables to the krr problem (alpha).
template <class M>
int MakePolyL2Weights(
               const vector<DenseMatrix<typename M::ElementType> > &kernels,
               const M &data,
               const DenseVector<typename M::ElementType> &labels,
               double kernel_reg,
    	       double alg_reg,
               double interp_param,
               double stop_tol,
               size_t max_iter,
               DenseVector<typename M::ElementType> *weights,
               DenseVector<typename M::ElementType> *krr_weights) {
  typedef typename M::ElementType T;
  CHECK(interp_param >= 0 && interp_param <= 1);

  size_t iter = 0;
  int return_flag = 0;
  double gap, obj, norm;
  double obj_old = 1e100;
  weights->Clear();
  if (kernels.size() > 0) 
    weights->Resize(kernels.size(), kernel_reg / sqrt(kernels.size()));
  else //if (kernels.size() == 0)
    weights->Resize(data.NumColumn(),
                    kernel_reg / sqrt(data.NumColumn()));

  DenseVector<T> grad, weights_old, diff;
  VLOG(2) << "computing gradient";
  while (iter < max_iter) {
    if (kernels.size() == 0)
      obj = QuadGradientRank1(data, *weights, labels, alg_reg, &grad,
                              krr_weights);
    else
      obj = QuadGradient(kernels, *weights, labels, alg_reg, &grad,
                         krr_weights);
    weights_old = *weights;

    // make scale of gradient similar to that of weight vector
    norm = kernel_reg / sqrt(DotProduct(grad, grad));
    ScalarProduct(norm, &grad);

    if (obj_old / obj < 1 - 1e-5) // error if increase in objective
      return_flag = 1;
    obj_old = obj;

    // take step
    for (size_t i = 0; i < weights->Size(); ++i) {
      weights->SetElement(i, interp_param * weights->Element(i)
                          + (1 - interp_param) * -grad.Element(i));
      if (weights->Element(i) < 0)
        LOG(WARNING) << "Negative weight " << i << ": " << weights->Element(i);
    }
    
    // project to feasible region,
    // maximum will always be at boundary since function is decreasing
    norm = sqrt(DotProduct(*weights, *weights));
    norm = kernel_reg / norm;
    ScalarProduct(norm, weights);

    Sum(-1.0, *weights, weights_old, &diff);
    gap = DotProduct(diff, diff);

    LOG(INFO) << "iter: " << iter << " obj: " << obj 
              //<< " gnorm: " << sqrt(DotProduct(grad, grad))
              //<< " wnorm: " << sqrt(DotProduct(*weights, *weights))
	      << " gap: " << gap;

    if (gap < stop_tol)
      break;
    ++iter;
  }

  return return_flag;
}

} //namespace kernel

#endif  // KERNEL_LIB_KERNEL_LEARNING_POLYL2_H_
