// kernel.cc
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Kernel lib definitions.


#include "kernel/lib/kernel.h"
#include "kernel/lib/feature-weighted-kernel.h"
#include "kernel/lib/rational-kernel.h"
#include "kernel/lib/composition-kernel.h"
#include "kernel/lib/product-kernel.h"
#include "kernel/lib/sum-kernel.h"
#include "kernel/lib/dense-vector.h"
#include "kernel/lib/sparse-vector.h"

namespace kernel {


// Define known object types
template <>
const string ObjectType<Fst<LogArc> >::type = "log_fst";
template <>
const string ObjectType<DenseVector<double> >::type = "double_dense_vector";
template <>
const string ObjectType<DenseVector<float> >::type = "float_dense_vector";
template <>
const string ObjectType<SparseVector<double> >::type = "double_sparse_vector";
template <>
const string ObjectType<SparseVector<float> >::type = "float_sparse_vector";

// Register Kernel classes for LogFst
REGISTER_KERNEL(RationalKernel, LogArc, "rational");
REGISTER_KERNEL(UnfactoredRationalKernel, LogArc, "unfactored_rational");
REGISTER_KERNEL(PolynomialKernel, LogFst, "polynomial");
REGISTER_KERNEL(SigmoidKernel, LogFst, "sigmoid");
REGISTER_KERNEL(GaussianKernel, LogFst, "gaussian");
REGISTER_KERNEL(ProductKernel, LogFst, "product");
REGISTER_KERNEL(SumKernel, LogFst, "sum");

// Register Kernel classes for SparseVector<double>
typedef SparseVector<double> DoubleSparseVector;
REGISTER_KERNEL(FeatureWeightedKernel, DoubleSparseVector, "feature_weighted");
REGISTER_KERNEL(PolynomialKernel, DoubleSparseVector, "polynomial");
REGISTER_KERNEL(SigmoidKernel, DoubleSparseVector, "sigmoid");
REGISTER_KERNEL(GaussianKernel, DoubleSparseVector, "gaussian");
REGISTER_KERNEL(ProductKernel, DoubleSparseVector, "product");
REGISTER_KERNEL(SumKernel, DoubleSparseVector, "sum");


// Identifies stream data as a kernel (and its endianity)
static const int32 kKernelMagicNumber = 2129983209;

// Check kernel magic number and read in Kernel header.
bool KernelHeader::Read(istream &strm, const string &source) {
  int32 magic_number = 0;
  ReadType(strm, &magic_number);
  if (magic_number != kKernelMagicNumber) {
    LOG(ERROR) << "KernelHeader::Read: Bad Kernel header: " << source;
    return false;
  }

  ReadType(strm, &kerneltype_);
  ReadType(strm, &objecttype_);
  ReadType(strm, &version_);
  ReadType(strm, &flags_);
  ReadType(strm, &properties_);
  if (!strm)
    LOG(ERROR) << "KernelHeader::Read: read failed: " << source;
  return strm;
}

// Write Kernel magic number and Kernel header.
bool KernelHeader::Write(ostream &strm, const string &source) const {
  WriteType(strm, kKernelMagicNumber);
  WriteType(strm, kerneltype_);
  WriteType(strm, objecttype_);
  WriteType(strm, version_);
  WriteType(strm, flags_);
  WriteType(strm, properties_);
  if (!strm)
    LOG(ERROR) << "KernelHeader::Write: write failed: " << source;
  return strm;
}

}  // namespace kernel
