
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#ifndef KERNEL_LIB_KERNEL_LEARNING_CORR_H_
#define KERNEL_LIB_KERNEL_LEARNING_CORR_H_

#include "kernel/lib/dense-matrix.h"
#include "kernel/lib/semidense-matrix.h"
#include "kernel/lib/dense-vector.h"
#include "kernel/lib/sparse-vector.h"

#include <vector>

namespace kernel {

typedef vector<vector<double> > Matrix;
typedef vector<double> Vector;

//
// Declarations
//

//TODO(rostami) Eventually needs to also return a kernel object.

// Given an m-by-m train and n-by-n test matrix, labels vector,
// regularization parameter and corr_weights vector which was produced
// by MakeCorrelationWeights, will output the n-by-m
// correlation-kernel matrix consisting of weighted dot products.
template<typename T, class M>
void MakeCorrelationKernel(const M &train,
                           const M &test,
                           const DenseVector<T> &corr_weights,
                           DenseMatrix<T> *k_matrix);

// Same arguments as above, but without the correlation weights. The
// correlation weights are computed each time the function is called.
template<typename T, class M>
void MakeCorrelationKernel(const M &train,
                           const M &test,
                           const DenseVector<T> &labels,
                           double ker_reg,
                           DenseMatrix<T> *k_matrix);

// The correlation-kernel produced by taking the pairwise dot-products
// all points in data matrix. The correlations are computed each time
// this function is called.
template<typename T, class M>
void MakeCorrelationKernel(const M &data,
                           const DenseVector<T> &labels,
                           double ker_reg,
                           DenseMatrix<T> *k_matrix);

// Given the data, kernel regularization parameter and unnormalized weight
// vector, returns a normalized version of the weights such that Tr[K] =
// ker_reg.
template <typename T, class M>
void NormalizeCorrelationWeights(const M &train,
                                 const DenseVector<T> &corr_weights,
                                 double ker_reg,
                                 DenseVector<T> *weights);


// Given a data matrix and label vector, computes a quantity
// proportional to the maginitude of the correlation of each feature
// with the labels.
template <typename T, class M>
void MakeCorrelationWeights(const M &data,
                            const DenseVector<T> &labels,
                            double ker_reg,
                            DenseVector<T> *weights);

// Same as above, but returns exactly the correlation keeping the
// sign as well as maginitude.
template <typename T, class M>
void MakeCorrelationWeights(const M &data,
                            const DenseVector<T> &labels,
                            DenseVector<T> *weights);

// Auxiliary function used for computing the (squared) norms of each
// column in the matrix mat.
template <typename T>
void ComputeSquaredNorms(const DenseMatrix<T> &mat, DenseVector<T> *norms);

template <typename T>
void ComputeSquaredNorms(const SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW> &mat,
                  DenseVector<T> *norms);

// Compute the sum of the elements of each column
//
template <typename T> inline
void ComputeSums(const DenseMatrix<T> &mat, DenseVector<T> *sums);


template <typename T> inline
void ComputeSums(const SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW> &mat,
                  DenseVector<T> *sums);

//
// Implementations
//

template<typename T, class M>
void MakeCorrelationFeatures(const M &data,
                             const DenseVector<T> &weights,
                             M *features) {
  DenseVector<T> weights_sqrt(weights.Size(), 0);
  for (size_t i = 0; i < weights_sqrt.Size(); ++i)
    weights_sqrt.SetElement(i, sqrt(weights.Element(i)));

  DiagonalProduct(data, weights_sqrt, features);
}

template<typename T, class M>
void MakeCorrelationKernel(const M &train,
                           const M &test,
                           const DenseVector<T> &weights,
                           DenseMatrix<T> *k_matrix) {
  CHECK(train.NumColumn() == test.NumColumn());

  M temp;
  typename M::TransposeMatrix diag;
  Diagonal(weights, &diag);

  Product(test, diag, &temp);

  typename M::TransposeMatrix train_t;
  train.Transpose(&train_t);
  Product(temp, train_t, k_matrix);
}

template<typename T, class M>
void MakeCorrelationKernel(const M &train,
                           const M &test,
                           const DenseVector<T> &labels,
                           double ker_reg,
                           DenseMatrix<T> *k_matrix) {
  DenseVector<T> weights(train.NumColumn(), 0);

  MakeCorrelationWeights(train, labels, ker_reg, &weights);

  MakeCorrelationKernel(train, test, weights, k_matrix);
}

template<typename T, class M>
void MakeCorrelationKernel(const M &data,
                           const DenseVector<T> &labels,
                           double ker_reg,
                           DenseMatrix<T> *k_matrix) {
  MakeCorrelationKernel(data, data, labels, ker_reg, k_matrix);
}

template <typename T, class M>
void NormalizeCorrelationWeights(const M &train,
                                 const DenseVector<T> &corr_weights,
                                 double ker_reg,
                                 DenseVector<T> *weights) {
  *weights = corr_weights;
  DenseVector<T> feature_norms(train.NumColumn(), 0);
  double norm_factor = 0;
  ComputeSquaredNorms(train, &feature_norms);

  for (size_t i = 0; i < weights->Size(); i++) {
    weights->SetElement(i, pow(weights->Element(i),2));
    norm_factor += feature_norms.Element(i) * weights->Element(i);
  }

  for (size_t i = 0; i < weights->Size(); i++) {
    weights->SetElement(i, weights->Element(i) * ker_reg / norm_factor);
  }
}


template <typename T, class M>
void MakeCorrelationWeights(const M &data,
                            const DenseVector<T> &labels,
                            double ker_reg,
                            bool offset,
                            DenseVector<T> *weights) {
  int i;
  DenseVector<T> labels_centered(labels.Size(), 0);
  DenseVector<T> v, w;
  DenseVector<T> feature_norms(data.NumColumn(), 0);
  DenseVector<T> feature_sums(data.NumColumn(), 0);
  typename M::TransposeMatrix data_t;
  data.Transpose(&data_t);
  if(weights->Size() != data.NumColumn())
    weights->Resize(data.NumColumn(), 0);

  double labels_sum = Sum(labels);
  for (size_t i = 0; i < labels.Size(); ++i)
    labels_centered.SetElement(i, labels.Element(i) - labels_sum / labels.Size());

  // (x - x_bar)' * y_centered = x' * y_centered - x_bar' * y_centered
  // v = data' * y_centered
  Product(data_t, labels_centered, &v);
  // w = x_bar' * y_centered = 0, calculation not needed!
  /*
  ComputeSums(data, &feature_sums);
  ScalarProduct(Sum(labels_centered) / data.NumRow(), feature_sums, &w);
  Sum(-1.0, w, &v);
  */

  ComputeCenteredNorms(data, &feature_norms);

  for (i = 0; i < v.Size(); i++) {
    if (feature_norms.Element(i) > 1e-10)
      weights->SetElement(i, v.Element(i) / feature_norms.Element(i));
    else
      weights->SetElement(i, 0);
  }

  if (offset) {
    // check if last element is a constant feature
    // b = Mean(y - Xw)
    if(feature_norms.Element(feature_norms.Size() - 1) > 1e-10) {
      LOG(INFO) << "Warning: last feature was not constant" << endl;
      weights->SetElement(feature_norms.Size() - 1, 0);
    }
    v.Clear();
    w.Clear();
    Product(data, *weights, &v);
    Sum(-1.0, v, labels, &w);
    weights->SetElement(feature_norms.Size() - 1, Mean(w));
  }

  if (ker_reg > 0) {
    DenseVector<T> weights_copy(*weights);
    NormalizeCorrelationWeights(data, weights_copy, ker_reg, weights);
  }
}

template <typename T, class M>
void MakeCorrelationWeights(const M &data,
                            const DenseVector<T> &labels,
                            double ker_reg,
                            DenseVector<T> *weights) {
  MakeCorrelationWeights(data, labels, ker_reg, false, weights);
}

template <typename T, class M>
void MakeCorrelationWeights(const M &data,
                            const DenseVector<T> &labels,
                            DenseVector<T> *weights) {
  MakeCorrelationWeights(data, labels, 0, weights);
}

template <typename T, class M> inline
void ComputeCenteredNorms(const M &mat, DenseVector<T> *norms) {
  // (x - x_bar)' * (x - x_bar) = x'*x - 2 x'*x_bar + x_bar'*x_bar

  // data_norms = x' * x
  DenseVector<T> data_norms, data_sums_sqr, offset;
  //DenseVector<T> xxbar, xbarxbar;
  ComputeSquaredNorms(mat, &data_norms);

  ComputeSums(mat, &data_sums_sqr);
  for (size_t i = 0; i < data_sums_sqr.Size(); ++i)
    data_sums_sqr.SetElement(i, pow(data_sums_sqr.Element(i), 2));

  // offset = - 2 x'*x_bar + x_bar'*x_bar = -1 x' * x_bar
  // Note: x_bar is a vector with all elements equal to 1/m sum_i(x_i)
  ScalarProduct(-1.0 / mat.NumRow(), data_sums_sqr, &offset);

  Sum(data_norms, offset, norms);
}

template <typename T> inline
void ComputeSums(const DenseMatrix<T> &mat, DenseVector<T> *sums) {
  sums->Clear();
  if (sums->Size() != mat.NumColumn())
    sums->Resize(mat.NumColumn(), 0);
  for (size_t i = 0; i < mat.NumRow(); ++i)
    for (size_t j = 0; j < mat.NumColumn(); ++j)
      sums->SetElement(j, sums->Element(j)
                               + mat.Element(i,j));
}

template <typename T> inline
void ComputeSums(const SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW> &mat,
                  DenseVector<T> *sums) {
  sums->Clear();
  if(sums->Size() != mat.NumColumn())
    sums->Resize(mat.NumColumn(), 0);

  SparseVector<T> row;
  pair<size_t, T> p;
  for (size_t i = 0; i < mat.NumRow(); ++i) {
    row = mat.Row(i);
    for( size_t j = 0; j < row.Size(); ++j) {
      p = row.Pair(j);
      sums->SetElement(p.first, sums->Element(p.first)
                        + p.second);
    }
  }
}

template <typename T> inline
void ComputeSquaredNorms(const DenseMatrix<T> &mat, DenseVector<T> *norms) {
  norms->Clear();
  if (norms->Size() != mat.NumColumn())
    norms->Resize(mat.NumColumn(), 0);
  for (size_t i = 0; i < mat.NumRow(); ++i)
    for (size_t j = 0; j < mat.NumColumn(); ++j)
      norms->SetElement(j, norms->Element(j)
                               + pow(mat.Element(i,j),2));
}

template <typename T> inline
void ComputeSquaredNorms(const SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW> &mat,
                  DenseVector<T> *norms) {
  norms->Clear();
  if(norms->Size() != mat.NumColumn())
    norms->Resize(mat.NumColumn(), 0);

  SparseVector<T> row;
  pair<size_t, T> p;
  for (size_t i = 0; i < mat.NumRow(); ++i) {
    row = mat.Row(i);
    for( size_t j = 0; j < row.Size(); ++j) {
      p = row.Pair(j);
      norms->SetElement(p.first, norms->Element(p.first)
                        + pow(p.second, 2));
    }
  }
}

} //namespace kernel

#endif  // KERNEL_LIB_KERNEL_LEARNING_CORR_H_
