// ngram-fst.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Representation of n-gram counting transducer.

#ifndef KERNEL_LIB_NGRAM_FST_H__
#define KERNEL_LIB_NGRAM_FST_H__

#include <fst/fstlib.h>

namespace kernel {

using namespace fst;

template <class A>
void Ngram(VectorFst<A> *fst, typename A::StateId n, typename A::Label sigma,
           typename A::StateId min_n = -1) {
  for (typename A::StateId i = 0; i <= n; ++i)
    fst->AddState();
  fst->SetStart(0);
  fst->SetFinal(n, A::Weight::One());

  for (typename A::Label l = 1; l <= sigma; ++l) {
    fst->AddArc(0, A(l, 0, A::Weight::One(), 0));
    fst->AddArc(n, A(l, 0, A::Weight::One(), n));
  }

  for (typename A::StateId i = 0; i < n; ++i)
    for (typename A::Label l = 1; l <= sigma; ++l)
      fst->AddArc(i, A(l, l, A::Weight::One(), i + 1));

  //  // Use epsilon-transitions to allow lower order ngrams,
  //  // conflicts with gappy n-gram handling.
  //  if (min_n >= 0)
  //  for (typename A::StateId i = min_n; i < n; ++i)
  //    fst->AddArc(i, A(0, 0, A::Weight::One(), n));

  if (min_n >= 0)
    for (typename A::StateId i = min_n - 1; i < n - 1; ++i)
      if (i < 0 )
        fst->AddArc(0, A(0, 0, A::Weight::One(), n));
      else
        for (typename A::Label l = 1; l <= sigma; ++l)
          fst->AddArc(i, A(l, l, A::Weight::One(), n));

  ArcSort(fst, ILabelCompare<A>());
}

template <class A>
void GappyNgram(VectorFst<A> *fst,
                typename A::StateId n,
                typename A::Label sigma,
                typename A::Weight lambda,
                typename A::StateId max = -1,
                typename A::StateId min_n = -1) {
  fst->DeleteStates();
  bool bounded = (max >= 0);
  VectorFst<A> *wfst = bounded ? new VectorFst<A>() : fst;
  Ngram(wfst, n, sigma, min_n);
  for (typename A::StateId i = 1; i < n; ++i)
    for (typename A::Label l = 1; l <= sigma; ++l)
      wfst->AddArc(i, A(l, bounded ? sigma + 1 : 0, lambda, i));

  if (bounded) {
    VectorFst<A> bfst;
    for (typename A::StateId i = 0; i <= max; ++i)
      bfst.SetFinal(bfst.AddState(), A::Weight::One());
    bfst.SetStart(0);
    for (typename A::StateId i = 0; i <= max; ++i) {
      for (typename A::Label l = 1; l <= sigma; ++l)
        bfst.AddArc(i, A(l, l, A::Weight::One(), i));
      if (i < max)
        bfst.AddArc(i, A(sigma + 1, 0, A::Weight::One(), i + 1));
    }
    ArcSort(wfst, OLabelCompare<A>());
    ArcSort(&bfst, ILabelCompare<A>());
    ComposeFst<A> cfst(*wfst, bfst);
    Map(cfst, fst, IdentityMapper<A>());
    delete wfst;
    if (min_n == -1) {
      EncodeMapper<A> encoder(kEncodeLabels | kEncodeWeights, ENCODE);
      Encode(fst, &encoder);
      Minimize(fst);
      Decode(fst, encoder);
    }
  }

  ArcSort(fst, ILabelCompare<A>());
}


}  // namespace kernel

#endif  // KERNEL_LIB_NGRAM_FST_H__
