// rational-function.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Rational kernel function representation.

#ifndef KERNEL_LIB_RATIONAL_FUNCTION_H__
#define KERNEL_LIB_RATIONAL_FUNCTION_H__

#include <fst/fstlib.h>
#include "kernel/lib/kernel-function.h"
#include "kernel/lib/kernel.h"

namespace kernel {

using namespace fst;

struct KernelReadOptions;
struct KernelWriteOptions;


enum EpsMapperType { EPS_ON_INPUT = 1, EPS_ON_OUTPUT = 2 };

template <class A, EpsMapperType T> class EpsMapper {
 public:
  explicit EpsMapper() {}

  A operator()(const A &arc) {
    return A(T == EPS_ON_INPUT ? 0 : arc.ilabel,
             T == EPS_ON_OUTPUT ? 0 : arc.olabel,
             arc.weight, arc.nextstate);
  }

  uint64 Properties(uint64 props) {
    uint64 oprops;
    if (T == EPS_ON_INPUT) {
      oprops = props & kILabelInvariantProperties;
      oprops |= kIEpsilons;
      if (props & kOEpsilons)
        oprops |= kEpsilons;
      if (props & kNoOEpsilons)
        oprops |= kNoEpsilons;
    } else {  // T == EPS_ON_OUTPUT
      oprops = props & kOLabelInvariantProperties;
      oprops |= kOEpsilons;
      if (props & kIEpsilons)
        oprops |= kEpsilons;
      if (props & kNoIEpsilons)
        oprops |= kNoEpsilons;
    }
    return oprops;
  }

  MapFinalAction FinalAction() const { return MAP_NO_SUPERFINAL; }

  MapSymbolsAction InputSymbolsAction() const { return MAP_COPY_SYMBOLS; }

  MapSymbolsAction OutputSymbolsAction() const { return MAP_COPY_SYMBOLS;}

};


template <class A>
class RationalFunction
    : public FeaturedFunction< Fst<A>, VectorFst<A> > {
 public:
  typedef Fst<A> Object;
  typedef VectorFst<A> Feature;

  RationalFunction(const Fst<A> &fst, int32 opt = -1)
      : fst_(fst.Copy()), opt_level_(opt) {
    if (!(fst_->Properties(kILabelSorted, true) & kILabelSorted))
      LOG(FATAL) << "RationalKernel: input fst not input-label sorted";
  }

  ~RationalFunction() {
    delete fst_;
  }

  Feature *GetFeature(const Fst<A>& x) const {
    VectorFst<A> *y =
      new VectorFst<A>(ProjectFst<A>(ComposeFst<A>(x, *fst_),
                                     PROJECT_OUTPUT));
    Connect(y);
    VectorFst<A> *y_opt;
    switch (y->Properties(kCyclic, true) ? 0 : opt_level_) {
      case 0: // none
        Map(y, EpsMapper<A, EPS_ON_INPUT>());
        ArcSort(y, OLabelCompare<A>());
        return y;
      case 1: // rmeps()
        RmEpsilon(y);
        Map(y, EpsMapper<A, EPS_ON_INPUT>());
        ArcSort(y, OLabelCompare<A>());
        return y;
      case 2: // det()
        y_opt = new VectorFst<A>(DeterminizeFst<A>(*y));
        Minimize(y_opt);
        delete y;
        Map(y_opt, EpsMapper<A, EPS_ON_INPUT>());
        ArcSort(y_opt, OLabelCompare<A>());
        return y_opt;
      case 3: // det(rmeps())
        RmEpsilon(y);
        y_opt = new VectorFst<A>(DeterminizeFst<A>(*y));
        Minimize(y_opt);
        delete y;
        Map(y_opt, EpsMapper<A, EPS_ON_INPUT>());
        ArcSort(y_opt, OLabelCompare<A>());
        return y_opt;
      case 4: // rmeps(det)
        y_opt = new VectorFst<A>(DeterminizeFst<A>(*y));
        Minimize(y_opt);
        delete y;
        RmEpsilon(y_opt);
        Map(y_opt, EpsMapper<A, EPS_ON_INPUT>());
        ArcSort(y_opt, OLabelCompare<A>());
        return y_opt;
      case 5: // det(rmeps(det()))
      case -1:
        if (FLAGS_v) cerr << "[" << y->NumStates() << ", ";
        y_opt = new VectorFst<A>(DeterminizeFst<A>(*y));
        if (FLAGS_v) cerr << "d " << y_opt->NumStates() <<", ";
        //Minimize(y_opt); //Bad because moves around weights
        // needs to be encoded.
        delete y;
        if (FLAGS_v) cerr << "m " << y_opt->NumStates() <<", ";
        RmEpsilon(y_opt);
        if (FLAGS_v) cerr << "r " << y_opt->NumStates() <<", ";
        y = new VectorFst<A>(DeterminizeFst<A>(*y_opt));
        if (FLAGS_v) cerr << "d " << y->NumStates() <<", ";
        Minimize(y);
        if (FLAGS_v) cerr << "m " << y->NumStates() <<"]";
        delete y_opt;
        Map(y, EpsMapper<A, EPS_ON_INPUT>());
        ArcSort(y, OLabelCompare<A>());
        return y;
      default: // none
        Map(y, EpsMapper<A, EPS_ON_INPUT>());
        ArcSort(y, OLabelCompare<A>());
        return y;
    }
  }

  double Product(const Feature& fx, const Feature& fy) const {
    VectorFst<A> ify(fy);
    if (FLAGS_v) cerr << "[" << ify.NumStates() << ",";
    Invert(&ify);
    CacheOptions opts = CacheOptions();
    opts.gc_limit = 0;
    ComposeFst<A> ccfst(fx, ify, ComposeFstOptions<A>(opts));
    DeterminizeFst<A> cfst(ccfst, DeterminizeFstOptions<A>(opts));
    VectorFst<A> vcfst(cfst);
    if (FLAGS_v) cerr << vcfst.NumStates() << ",";
    Connect(&vcfst);
    if (FLAGS_v) cerr << vcfst.NumStates() << ",";
    vector<typename A::Weight> distance;
    ShortestDistance(vcfst, &distance, true);
    double value = vcfst.Start() != -1 && vcfst.Start() < distance.size()
                   ? exp(-distance[vcfst.Start()].Value()) : 1.0;
    if (FLAGS_v && vcfst.Start() != -1 && vcfst.Start() < distance.size())
      cerr << "(" << vcfst.Start() << ","
           << distance.size() << ":" << distance[vcfst.Start()]
           << ")";
    if (FLAGS_v)
      cerr << value << "]";
    return value != 0.0 ? value : 1.0;
  }

  double operator()(const Fst<A>& x, const Fst<A>& y) const {
    VectorFst<A> *fx, *fy;
    fx = GetFeature(x);
    fy = GetFeature(y);
    double value = Product(*fx, *fy);
    delete fx;
    delete fy;
    return value;
  }

  const string Type() const { return type_;}

  const uint64 Properties() const { return properties_;}

  static RationalFunction<A> *Read(istream &strm,
                                     const KernelReadOptions &opts) {
    int32 opt_level;
    ReadType(strm, &opt_level);
    Fst<A> *fst = Fst<A>::Read(strm, FstReadOptions());
    if (!fst) {
      LOG(ERROR) << "RationalFunction::Read: read failed: "
                 << opts.source;
      return 0;
    }
    RationalFunction<A> *function = new RationalFunction<A>(*fst, opt_level);
    delete fst;
    return function;
  }

  bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    WriteType(strm, opt_level_);
    return fst_->Write(strm, FstWriteOptions());
  }

 private:
  Fst<A> *fst_;
  int32 opt_level_;
  static const string type_;
  static const uint64 properties_ = kFeatured | kSymmetric | kPosDefinite;
};

template <class A>
const string RationalFunction<A>::type_ = "rational";


template <class A>
class UnfactoredRationalFunction
    : public FeaturedFunction< Fst<A>, VectorFst<A> > {
 public:
  typedef Fst<A> Object;
  typedef VectorFst<A> Feature;

  UnfactoredRationalFunction(const Fst<A>& fst, uint32 opt_level = -1) :
      fst_(fst.Copy()), opt_level_(opt_level) { }


  ~UnfactoredRationalFunction() {
    delete fst_;
  }

  Feature *GetFeature(const Fst<A>& x) const {
    return new VectorFst<A>(x);
  }

  double Product(const Feature& fx, const Feature& fy) const {
    return operator()(fx, fy);
  }

  double operator()(const Fst<A>& x, const Fst<A>& y) const {
    switch (opt_level_) {
      default:
      case 3: {
        CacheOptions opts = CacheOptions();
        opts.gc_limit = 0;

        VectorFst<A> vx, vy;
        Map(x, &vx, EpsMapper<A, EPS_ON_INPUT>());
        Map(y, &vy, EpsMapper<A, EPS_ON_OUTPUT>());

        ComposeFst<A> ccfst(ComposeFst<A>(vx, *fst_,
                                         ComposeFstOptions<A>(opts)),
                           vy, ComposeFstOptions<A>(opts));

        DeterminizeFst<A> cfst(ccfst, DeterminizeFstOptions<A>(opts));
        VectorFst<A> fst(cfst);
        if (FLAGS_v) cerr << "[2w " << fst.NumStates() << ",";
        Connect(&fst);
        if (FLAGS_v) cerr << fst.NumStates() << ":";
        vector<typename A::Weight> distance;
        ShortestDistance(fst, &distance, true);
        double value = fst.Start() != -1 && fst.Start() < distance.size()
                   ? exp(-distance[fst.Start()].Value()) : 1.0;
        if (FLAGS_v) cerr << value << "]";
        return value != 0.0 ? value : 1.0;

      }

      case -1:
      case 0: {
        VectorFst<A> vx, vy;
        Map(x, &vx, EpsMapper<A, EPS_ON_INPUT>());
        Map(y, &vy, EpsMapper<A, EPS_ON_OUTPUT>());

        VectorFst<A> c1(ComposeFst<A>(vx, *fst_));
        if (FLAGS_v) cerr << "[" << c1.NumStates() << ",";
        Connect(&c1);
        if (FLAGS_v) cerr << c1.NumStates() << ",";
        ArcSort(&c1, OLabelCompare<A>());

        CacheOptions opts = CacheOptions();
        opts.gc_limit = 0;
        VectorFst<A>
          c2(DeterminizeFst<A>(ComposeFst<A>(c1, vy,
                                             ComposeFstOptions<A>(opts)),
                               DeterminizeFstOptions<A>(opts)));

        if (FLAGS_v) cerr << c2.NumStates() << ",";
        Connect(&c2);
        if (FLAGS_v) cerr << c2.NumStates() << ":";
        vector<typename A::Weight> distance;
        ShortestDistance(c2, &distance, true);
        double value = c2.Start() != -1 && c2.Start() < distance.size()
                   ? exp(-distance[c2.Start()].Value()) : 1.0;
        if (FLAGS_v) cerr << value << "]";
        return value != 0.0 ? value : 1.0;
      }
    }
  }

  const string Type() const { return type_;}

  const uint64 Properties() const { return properties_;}

  static UnfactoredRationalFunction<A> *Read(istream &strm,
                                             const KernelReadOptions &opts) {
    int32 opt_level;
    ReadType(strm, &opt_level);
    Fst<A> *fst = Fst<A>::Read(strm, FstReadOptions());
    if (!fst) {
      LOG(ERROR) << "UnfactoredRationalFunction::Read: read failed: "
                 << opts.source;
      return 0;
    }
    UnfactoredRationalFunction<A> *function =
      new UnfactoredRationalFunction<A>(*fst, opt_level);
    delete fst;
    return function;
  }

  bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    WriteType(strm, opt_level_);
    return fst_->Write(strm, FstWriteOptions());
  }

 private:
  const Fst<A> *fst_;
  uint32 opt_level_;
  static const string type_;
  static const uint64 properties_ = 0;
};

template <class A>
const string UnfactoredRationalFunction<A>::type_ = "unfactored_rational";

}  // namespace kernel


#endif  // KERNEL_LIB_RATIONAL_FUNCTION_H__
