// register.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Classes required for registering kernel classes for i/o.

#ifndef KERNEL_LIB_REGISTER_H__
#define KERNEL_LIB_REGISTER_H__

#include <map>
#include <pthread.h>

#include "kernel/lib/compat.h"

#include "kernel/lib/object.h"
#include <iostream>
#include <fstream>

extern "C" {
  typedef void (*KernelInitFunc)(); // will only be used for dl monkey business
}

namespace kernel {

class KernelBase;
template <class O> class Kernel;
struct KernelReadOptions;


class KernelRegister {
 public:
  typedef KernelBase *(*Reader)(istream &strm, const KernelReadOptions &opts);

  static KernelRegister *GetRegister(){
    pthread_once(&register_init_, &KernelRegister::Init);
    return register_;
  }

  void SetReader(const  string &object_type, const string &kernel_type,
                 const Reader &reader) {
    MutexLock l(register_lock_);
    kernel_table_.insert(make_pair(make_pair(object_type, kernel_type), reader));
  }

  const Reader GetReader(const string &object_type,
                         const string &kernel_type) const {
    Reader reader = LookupReader(object_type, kernel_type);
    if(!reader)
      LOG(ERROR) << "KernelRegister::GetReader: unknown type ("
                 << object_type << ", " << kernel_type << ")";
    return reader;
    // TODO: add dl monkey business
  }

 private:
  static void Init() {
    register_lock_ = new Mutex;
    register_ = new KernelRegister;
  }

  Reader LookupReader(const string &object_type,
                         const string &kernel_type) const {
    MutexLock l(register_lock_);
    map< pair<string, string>, Reader>::const_iterator it =
      kernel_table_.find(make_pair(object_type, kernel_type));
    if (it != kernel_table_.end())
      return it->second;
    else
      return 0;
  }

  static pthread_once_t register_init_;  // ensures only called once
  static Mutex* register_lock_;          // multithreading lock
  static KernelRegister *register_;

  map<pair<string, string>, Reader> kernel_table_;
};

// This class registers a Kernel type for generic reading and creating.
template <class K>
class KernelRegisterer {
 public:
  typedef typename K::Object Object;
  typedef typename KernelRegister::Reader Reader;

  KernelRegisterer(const string &type) {
    K *(*reader)(istream &strm,
                 const KernelReadOptions &opts) = &K::Read;

    KernelRegister *registr = KernelRegister::GetRegister();
    registr->SetReader(ObjectType<Object>::type, type,
                      reinterpret_cast<Reader>(reader));
  }
};


// Convenience macro to generate static KernelRegisterer instance.
#define REGISTER_KERNEL(K, A, type)                                          \
  static kernel::KernelRegisterer< K <A> > K ## _ ## A ## _registerer(type)


class DataSetBase;

// This class defines mapping between dataset file names and
// the dataset objects.
class DataRegister {
 public:
  static DataRegister *GetRegister(){
    pthread_once(&register_init_, &DataRegister::Init);
    return register_;
  }

  void SetData(const  string &filename, DataSetBase *data) {
    MutexLock l(register_lock_);
    data_table_.insert(make_pair(filename, data));
  }

  DataSetBase *GetData(const string &filename) const {
    DataSetBase *data = LookupData(filename);
    return data;
  }

 private:
  static void Init() {
    register_lock_ = new Mutex;
    register_ = new DataRegister;
  }

  DataSetBase *LookupData(const string &filename) const {
    MutexLock l(register_lock_);
    map<string, DataSetBase*>::const_iterator it =
      data_table_.find(filename);
    if (it != data_table_.end())
      return it->second;
    else
      return 0;
  }

  static pthread_once_t register_init_;  // ensures only called once
  static Mutex* register_lock_;          // multithreading lock
  static DataRegister *register_;

  map<string, DataSetBase*> data_table_;
};


}  // namespace kernel

#endif  // KERNEL_LIB_REGISTER_H__
