#!/bin/bash

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: rostami@google.com (Afshin Rostamizadeh)

#set -x

# Paths to binaries, reset as needed.
BIN=~/src/openkernel/kernel/bin
SCRIPTS=~/src/openkernel/kernel/tools
LIBSVM=~/src/openkernel/libsvm-2.82
LIBLIN=~/liblinear-1.32

DATAFILE=$1
NUMTRIALS=$2
KER_ALG=$3
KER_RANGE=$4
KER_FLAGS=$5
LER_ALG=$6
LER_RANGE=$7
LER_TRAIN_FLAGS=$8
LER_PREDICT_FLAGS=$9

if [ $# -lt 9 ]; then
  echo Usage: $0 data_file num_trials kernel_alg kernel_range kernel_flags \
  learning_alg learning_range learning_train_flags learning_predict_flags
  exit 1
fi

echo $0 $1 $2 $3 $4 $5 $6 $7 $8 $9 > command.txt

if [ $KER_ALG = 'klweightfeatures' ]; then
  NUM_POINTS=`wc -l $DATAFILE | awk '{print $1}'`
else
  DATAFILE0=`head -n1 $DATAFILE`
  NUM_POINTS=`wc -l $DATAFILE0 | awk '{print $1}'`
fi

NUM_TEST=`echo "$NUM_POINTS * 1 / $NUMTRIALS" | bc`
NUM_TRAIN=`expr $NUM_POINTS - $NUM_TEST`

rm test_summary
rm train_summary

for ap in $LER_RANGE; do
  ALGPARAM=`echo "scale=5; 2^$ap" | bc`
  echo alg_param: $ALGPARAM

  for kp in $KER_RANGE; do
    KERPARAM=`echo "scale=5; 2^$kp" | bc`
    echo ker_param: $KERPARAM
    BASENAME=kp-$KERPARAM\_ap-$ALGPARAM.err
    rm *$BASENAME

    TRAIN_START=$NUM_TEST
    for trial in `seq 1 $NUMTRIALS`; do
      $BIN/$KER_ALG $KER_FLAGS \
               --num_train=$TRAIN_START:`expr $TRAIN_START + $NUM_TRAIN` \
               --alg_reg=$ALGPARAM --ker_reg=$KERPARAM $DATAFILE kernel.libsvm

      split -l$NUM_TEST -d kernel.libsvm fold
      TEST_FOLD_NUM=`expr $trial - 1`
      mv fold`printf "%.2d" $TEST_FOLD_NUM` kernel_test.libsvm
      cat fold* > kernel_train.libsvm
      rm fold*

      if [ $LER_ALG = 'krr' ]; then
        $BIN/krr-train $LER_TRAIN_FLAGS kernel_train.libsvm $ALGPARAM model
        $BIN/krr-predict $LER_PREDICT_FLAGS kernel_test.libsvm model \
          &> test_err
        $BIN/krr-predict $LER_PREDICT_FLAGS kernel_train.libsvm model \
          &> train_err
        TESTERR=`grep RMSE test_err | cut -f3 -d':'`
        TRAINERR=`grep RMSE train_err | cut -f3 -d':'`
      fi
      
      if [ $LER_ALG = 'svm' ]; then
        $LIBSVM/svm-train $LER_TRAIN_FLAGS -c $ALGPARAM kernel_train.libsvm \
           model
        $LIBSVM/svm-predict $LER_PREDICT_FLAGS kernel_test.libsvm model \
           kernel_test.libsvm.pred > test_err
        $LIBSVM/svm-predict $LER_PREDICT_FLAGS kernel_train.libsvm model \
           kernel_train.libsvm.pred > train_err
        TESTACC=`grep Acc test_err | cut -f2 -d'=' | cut -f1 -d'%'`
        TRAINACC=`grep Acc train_err | cut -f2 -d'=' | cut -f1 -d'%'`
        TESTERR=`echo 100 - $TESTACC | bc`
        TRAINERR=`echo 100 - $TRAINACC | bc`
      fi

      if [ $LER_ALG = 'lin' ]; then
        $LIBLIN/train $LER_TRAIN_FLAGS -c $ALGPARAM kernel_train.libsvm model
        $LIBLIN/predict $LER_PREDICT_FLAGS kernel_test.libsvm model \
          kernel_test.libsvm.pred > test_err
        $LIBLIN/predict $LER_PREDICT_FLAGS kernel_train.libsvm model \
          kernel_train.libsvm.pred > train_err
        TESTACC=`grep Acc test_err | cut -f2 -d'=' | cut -f1 -d'%'`
        TRAINACC=`grep Acc train_err | cut -f2 -d'=' | cut -f1 -d'%'`
        TESTERR=`echo 100 - $TESTACC | bc`
        TRAINERR=`echo 100 - $TRAINACC | bc`
      fi

      echo $TESTERR >> test_err_$BASENAME
      echo $TRAINERR >> train_err_$BASENAME
      printf "trial: %d ap: %g kp: %g test_err: %g train_err: %g\n" \
      $trial $ALGPARAM $KERPARAM $TESTERR $TRAINERR

      rm kernel.libsvm kernel_train.libsvm kernel_test.libsvm
      #cycle data file
      TRAIN_START=`expr $TRAIN_START + $NUM_TEST`
    done #for trial

    echo Calculating stats...
    $SCRIPTS/stats.py test_err_$BASENAME >> test_summary
    $SCRIPTS/stats.py train_err_$BASENAME >> train_summary
  done #for kp
done #for ap

rm model
rm kernel_test.libsvm.pred   
rm kernel_train.libsvm.pred   
rm test_err train_err
rm *.err #comment this line in order to keep per trial stats
